import pytest




from web3.utils.datastructures import (
    AttributeDict,
)


@pytest.mark.asyncio
async def test_some_asyncio_code():
    from ethexporter import context
    await context.initial()
    from pprint import pprint
    v = await context['lastblock']
    print(v)
    pprint(AttributeDict.recursive(v[0].value))



from collections import defaultdict
import logging
from aioethereum import create_ethereum_client

_log = logging.getLogger(__name__)
async def nodes(context):
    from ethexporter import LabeledValue as LV
    for node, url in (await context['urls']).items():
        yield LV(await create_ethereum_client(url), {'node': node, 'url': url})


async def eth_lastblock_num(context):
    from ethexporter import LabeledValue as LV
    for node in await context['nodes']:
        try:
            yield LV(await node.value.eth_blockNumber(), node.labels)
        except ConnectionError as e:
            _log.exception(e)


async def lastblock(context):
    from ethexporter import LabeledValue as LV
    nodes = await context['nodes']
    lastblocks = await context['eth_lastblock_num']
    node_lastblocks = defaultdict(list)
    for node in nodes:
        for lastblock in lastblocks:
            if lastblock.labels['node'] == node.labels['node']:
                node_lastblocks[node.labels['node']].extend([node.value, lastblock.value, lastblock.labels])

    for name, (node, lastblock, labels) in node_lastblocks.items():
        try:
            yield LV(await node.eth_getBlockByNumber(lastblock), labels)
        except ConnectionError as e:
            _log.exception(e)

async def eth_lastblock_size(context):
    from ethexporter import LabeledValue as LV
    for lastblock in await context['lastblock']:
        yield LV(int(lastblock.value['size'], 0), lastblock.labels)


async def eth_gas_limit(context):
    from ethexporter import LabeledValue as LV
    for lastblock in await context['lastblock']:
        yield LV(int(lastblock.value['gasLimit'], 0), lastblock.labels)


async def eth_gas_used(context):
    from ethexporter import LabeledValue as LV
    for lastblock in await context['lastblock']:
        yield LV(int(lastblock.value['gasUsed'], 0), lastblock.labels)


import pytest
from ethexporter import ContextMapping
from random import random
import logging
_log = logging.getLogger(__name__)


@pytest.mark.asyncio
async def test_contextmapping_raises_key_error():
    m = ContextMapping(source={})
    with pytest.raises(KeyError):
        await m['xxx']


@pytest.mark.asyncio
async def test_contextmapping_caches_async_generator_once():
    async def _callable(context):
        for i in range(10):
            yield random() #mock.Mock(spec=) is broken on pytest-asyncio
    m = ContextMapping(source={'_callable1': _callable}, initial_keys=[])
    assert await m['_callable1'] == await m['_callable1']
    assert isinstance(await m['_callable1'], list)


@pytest.mark.asyncio
async def test_contextmapping_caches_awaitable_once():
    async def _callable(context):
        return random() #mock.Mock(spec=) is broken on pytest-asyncio
    m = ContextMapping(source={'_callable2': _callable}, initial_keys=[])
    assert await m['_callable2'] == await m['_callable2']
    assert isinstance(await m['_callable2'], float)


@pytest.mark.asyncio
async def test_contextmapping_caches_generator_once():
    def _callable(context):
        for i in range(10):
            yield random() #mock.Mock(spec=) is broken on pytest-asyncio
    m = ContextMapping(source={'_callable3': _callable}, initial_keys=[])
    assert await m['_callable3'] == await m['_callable3']
    assert isinstance(await m['_callable3'], list)


@pytest.mark.asyncio
async def test_contextmapping_caches_callable_once():
    def _callable(context):
        return random() #mock.Mock(spec=) is broken on pytest-asyncio
    m = ContextMapping(source={'_callable4': _callable}, initial_keys=[])
    assert await m['_callable4'] == await m['_callable4']
    assert isinstance(await m['_callable4'], float)


@pytest.mark.asyncio
async def test_contextmapping_len():
    assert len(ContextMapping(source={'a':'b'}, initial_keys=[])) == 1


@pytest.mark.asyncio
async def test_contextmapping_iter():
    assert list(ContextMapping(source={'a':'b'}, initial_keys=[])) == ['a',]


import time
from prometheus_aioexporter.script import PrometheusExporterScript
from prometheus_aioexporter.metric import MetricConfig
from functools import lru_cache
import asyncio
import prometheus_client.core
import logging
import sys
from toolrack.log import setup_logger

from . import config, context


_log = logging.getLogger(__name__)


class Web3Exporter(PrometheusExporterScript):
    """WEB3 RPC exporter."""


    def configure_argument_parser(self, parser):
        # Additional arguments to the script
        parser.add_argument('--rpc', nargs='*', default=None, help='Web3 RPC url(s)')
        parser.add_argument('--nodes', nargs='*', default=[], help='Name(s) for RPC url(s)')
        parser.add_argument('--gather', nargs='*', default=None, help=f'Limit gathered metrics to {list(config.metrics.keys())!r}')


    def configure(self, args):
        if args.rpc is not None:
            if len(args.nodes) < len(args.rpc):
                for i in range(len(args.rpc) - len(args.nodes)):
                    args.nodes.append(f'node{i}')
            config.initial['urls'] = dict(zip(args.nodes[:len(args.rpc)], args.rpc))
            context._source['urls'] = config.initial['urls']

        declarations = []
        for name, declaration in self.metrics:
            declarations.append(MetricConfig(name,
                                        declaration['description'],
                                        declaration['type'],
                                        declaration['options']))
        self.create_metrics(declarations)
        self._last_request_time = 0

        if args.gather:
            config.gather = args.gather
        gather = getattr(config, 'gather', None)
        if gather is not None:
            ungather = [metric_name for metric_name in self.registry._metrics if metric_name not in gather]
            for metric_name in ungather:
                del self.registry._metrics[metric_name]
        _log.info('gathering', extra={'metrics': list(self.registry._metrics)})

    async def on_application_startup(self, application):
        await context.initial()
        application['exporter'].set_metric_update_handler(self._update_handler)

    def _iter_metrics(self):
        _metrics = getattr(config, 'metrics', {})
        for name in sorted(_metrics):
            yield name, _metrics[name]


    @property
    @lru_cache(maxsize=1)
    def metrics(self):
        return list(self._iter_metrics())


    async def _update_handler(self, application):
        # Stop other asyncio tasks at application shutdown
        _request_delta = time.time() - self._last_request_time
        if  _request_delta > config.scrape_throttle:
            _log.debug('args', extra=application)
            await self.scrape_metrics(application)
            context.reset() #contextvar support in asyncio added only in py37
            self._last_request_time = time.time()
        else:
            _log.debug(f'throttling after {_request_delta!r}')



    async def scrape_metrics(self, application):
        for name, declaration in self.metrics:
            if name not in application:
                continue
            _log.debug(f'{name!r}')
            metric = application[name]
            for labeled in await context[name]:
                if isinstance(metric, prometheus_client.core._LabelWrapper):
                    label_formats = declaration.get('options', {}).get('labels', {})
                    labels = dict((k, v.format(**labeled.labels)) for (k, v) in label_formats.items())
                    _metric = metric.labels(**dict(labels))
                else:
                    _metric = metric
                    labels = {}
                declaration['setter'](_metric)(labeled.value)
                _log.debug('exposed', extra={'key':name, 'value': labeled.value, 'labels':labels})

    def _setup_logging(self, log_level):
        """Setup logging for the application and aiohttp."""
        level = getattr(logging, log_level)
        names = (
            'aiohttp.access', 'aiohttp.internal', 'aiohttp.server',
            'aiohttp.web', 'aiohttp.client', 'asyncio_client', self.name)
        for name in names:
            setup_logger(name=name, stream=sys.stderr, level=level)

exporter = Web3Exporter()


import devconfig
devconfig.merges['ethexporter_config']._with('ethexporter_config', 'ethexporter.defaults')

try:
    import ethexporter_config as config
except ModuleNotFoundError:
    import ethexporter.defaults as config

import inspect
import logging
from collections import namedtuple
import collections.abc

import devconfig.mapping

from ethexporter import config
_log = logging.getLogger(__name__)

LabeledValue = namedtuple('Labeled', ('value', 'labels'))


_ctx_source = devconfig.mapping.merge(
                dict((name, metric['scraper']) for name, metric in getattr(config, 'metrics', {}).items()),
                getattr(config, 'scrapers', {}),
                getattr(config, 'initial', {}))


class ContextMapping(collections.abc.Mapping):
    def __init__(self, source=_ctx_source, initial_keys=getattr(config, 'initial', {}).keys()):
        self._source = source
        self._data = {}
        self._log = logging.getLogger(__name__ + '.context')
        self.initial_keys = initial_keys

    async def initial(self):
        [await self[k] for k in self.initial_keys]

    def __iter__(self):
        yield from self._source

    def __len__(self):
        return len(self._source)

    async def __getitem__(self, name):
        if name not in self._source:
            raise KeyError(name)

        if name not in self._data:
            logdata = {'context': self, 'key': name}
            scraper = self._source[name]
            if inspect.isasyncgenfunction(scraper):
                items = []
                self._log.debug(f'caching {name!r} key with async for', extra=logdata)
                async for item in scraper(self):
                    items.append(item)
            elif inspect.iscoroutinefunction(scraper):
                self._log.debug(f'caching {name!r} key with await', extra=logdata)
                items = await scraper(self)
            elif inspect.isgeneratorfunction(scraper):
                self._log.debug(f'caching {name!r} key with list', extra=logdata)
                items = list(scraper(self)) 
            elif inspect.isfunction(scraper):
                self._log.debug(f'caching {name!r} key with call', extra=logdata)
                items = scraper(self)
            else:
                self._log.debug(f'caching {name!r} key as value', extra=logdata)
                items = scraper
            self._data[name] = items

        return self._data[name]

    def reset(self):
        for k in self:
            if k not in self.initial_keys and k in self._data:
                _log.debug(f'removing non initial {k!r}')
                del self._data[k]


context = ContextMapping()